from abc import ABC, abstractmethod

from astropy.time import Time
from scipy.spatial.transform import Rotation


class BaseAttitudeStrategy(ABC):
    @abstractmethod
    def at(self, time: Time) -> Rotation:
        """
        Return the attitude Rotation instance at the specific time(s).
        By default, the attitude class has no reference frame attached to it,
        so it is up to the user to understand what reference frame their attitudes
        utilize.

        Args:
            time (Time): Times to retrieve attitude data for.

        Returns:
            Rotation: scipy.spatial.transform.Rotation instance defining the attitude.
        """
        pass



from abc import ABC, abstractmethod

import numpy as np
from astropy.time import Time


class BaseAccessConstraint(ABC):
    def __init__(self) -> None:
        """
        Represents an access constraint which must be satisfied when provided to
        and access calcualtion.
        """
        super().__init__()

    @abstractmethod
    def __call__(self, time: Time) -> np.ndarray:
        """
        Method called by the Access algorithm which passes in a

        Args:
            time (Time): Astropy Time(s) to check this access constraint during.

        Returns:
            np.ndarray: boolean array matching the shape of the Time argument.
            True means access was successful at that time, False means access
            failed at that time.
        """
        pass



import csv
import os

import astropy.units as u
import numpy as np
from astropy.time import Time

from skywatch.skypath import SkyPath

LEO_EPHEM_CSV = os.path.join(os.path.dirname(__file__), "data", "leo_sat_ephem.csv")


def read_csv(path: str) -> dict:
    # Open the CSV file in read mode
    with open(path, "r") as file:
        # Create a CSV reader object
        csv_reader = csv.reader(file)

        # Iterate over each row in the CSV file
        rows = []
        for row in csv_reader:
            # Each row is a list of values representing columns
            rows.append(row)
    return rows


def get_ephem_data(path: str = LEO_EPHEM_CSV) -> dict:
    data = {}
    csv_data = read_csv(path)
    header = csv_data[0]

    column_data = [[] for _ in header]
    for row in csv_data[1:]:
        for index, column in enumerate(row):
            column_data[index].append(column)

    for heading, column in zip(header, column_data):
        data[heading] = column

    return data


def get_ephem_as_skypath(path: str = LEO_EPHEM_CSV) -> SkyPath:
    leo_csv = get_ephem_data(path)
    leo_csv_times = Time(leo_csv["utc"])
    leo_csv_position = (
        np.array(
            [leo_csv["x_eci_km"], leo_csv["y_eci_km"], leo_csv["z_eci_km"]]
        ).astype(float)
        * u.km
    )
    leo_csv_velocities = np.array(
        [leo_csv["vx_eci_km_s"], leo_csv["vy_eci_km_s"], leo_csv["vz_eci_km_s"]]
    ).astype(float) * (u.km / u.s)

    sat_position = SkyPath.from_ECI(
        leo_csv_times, *leo_csv_position, leo_csv_velocities
    )
    return sat_position



from astropy.time import Time
from scipy.spatial.transform import Rotation, Slerp

from skywatch.attitude.base_attitude import BaseAttitudeStrategy


class Slerped(BaseAttitudeStrategy):
    def __init__(self, times: Time, rotations: Rotation) -> None:
        """
        Using known attitudes (usually from real observations), do
        spherical linear interpolation between them.

        Args:
            times (Time): Times of known observations
            rotations (Rotation): Known attitudes
        """
        self.slerper = Slerp(times, rotations)

    def at(self, time: Time) -> Rotation:
        return self.slerper(time)



import math
from typing import List, Tuple

import astropy.units as u
import numpy as np


class PointGenerator:
    def __call__(self) -> List[Tuple[float, float]]: ...


class FibonacciPointGenerator(PointGenerator):
    def __init__(self, samples: int = 1000):
        self.samples = samples

    def __call__(self) -> List[tuple]:
        points = []
        phi = math.pi * (3.0 - math.sqrt(5.0))  # golden angle in radians

        for i in range(self.samples):
            y = 1 - (i / float(self.samples - 1)) * 2  # y goes from 1 to -1
            radius = math.sqrt(1 - y * y)  # radius at y

            theta = phi * i  # golden angle increment

            x = math.cos(theta) * radius
            z = math.sin(theta) * radius

            # Convert Cartesian coordinates to latitude and longitude
            lat = math.asin(y) * 180 / math.pi
            lon = math.atan2(z, x) * 180 / math.pi

            points.append((lat, lon))

        return points


def lat_lon_to_xyz(latitude, longitude, radius=6371):
    """
    Convert latitude and longitude to Cartesian coordinates.

    Parameters:
    latitude (float): Latitude in degrees.
    longitude (float): Longitude in degrees.
    radius (float): Radius of the sphere. Defaults to Earth's radius (6371 km).

    Returns:
    tuple: A tuple representing the Cartesian coordinates (X, Y, Z).
    """
    # Convert latitude and longitude from degrees to radians
    lat_rad = math.radians(latitude)
    lon_rad = math.radians(longitude)

    # Calculate Cartesian coordinates
    x = radius * math.cos(lat_rad) * math.cos(lon_rad)
    y = radius * math.cos(lat_rad) * math.sin(lon_rad)
    z = radius * math.sin(lat_rad)

    return x, y, z


@u.quantity_input(altitude=u.km, body_radius=u.km)
def max_elevation_angle(altitude: u.km, body_radius: u.km = 6371 * u.km):
    """Calculates the maximum elevation angle from NADIR
    a satellite can point to and still intersect the earth.
    """
    psi = np.arccos(body_radius / (body_radius + altitude))  # Horizon angle in radians
    e = (np.pi / 2) * u.rad - psi  # Maximum elevation angle in radians
    return np.degrees(e)  # Convert to degrees


# https://stephenhartzell.medium.com/satellite-line-of-sight-intersection-with-earth-d786b4a6a9b6
def los_to_earth(position, pointing):
    """Find the intersection of a pointing vector with the Earth
    Finds the intersection of a pointing vector u and starting point s with the WGS-84 geoid
    Args:
        position (np.array): length 3 array defining the starting point location(s) in meters
        pointing (np.array): length 3 array defining the pointing vector(s) (must be a unit vector)
    Returns:
        np.array: length 3 defining the point(s) of intersection with the surface of the Earth in meters

    Usage:
        To get the ground track of a satellite in cartesian coordinates, you would do the following calculation:
        ```
        los_to_earth(sat_pos_ecef.T.to(u.m).value, -sat_pos_ecef.T / np.linalg.norm(sat_pos_ecef.T))
        ```
        This gets the unit vector in the Z (nadir) direction for an ECEF position vector, and returns
        the position on the surface of the earth (in cartesian coords) that intersects the Z axis.
    """

    a = 6378137.0
    b = 6378137.0
    c = 6356752.314245
    x = position[0]
    y = position[1]
    z = position[2]
    u = pointing[0]
    v = pointing[1]
    w = pointing[2]

    value = -(a**2) * b**2 * w * z - a**2 * c**2 * v * y - b**2 * c**2 * u * x
    radical = (
        a**2 * b**2 * w**2
        + a**2 * c**2 * v**2
        - a**2 * v**2 * z**2
        + 2 * a**2 * v * w * y * z
        - a**2 * w**2 * y**2
        + b**2 * c**2 * u**2
        - b**2 * u**2 * z**2
        + 2 * b**2 * u * w * x * z
        - b**2 * w**2 * x**2
        - c**2 * u**2 * y**2
        + 2 * c**2 * u * v * x * y
        - c**2 * v**2 * x**2
    )
    magnitude = a**2 * b**2 * w**2 + a**2 * c**2 * v**2 + b**2 * c**2 * u**2

    if radical < 0:
        raise ValueError("The Line-of-Sight vector does not point toward the Earth")
    d = (value - a * b * c * np.sqrt(radical)) / magnitude

    if d < 0:
        raise ValueError("The Line-of-Sight vector does not point toward the Earth")

    return np.array(
        [
            x + d * u,
            y + d * v,
            z + d * w,
        ]
    )



import datetime
import time
import unittest

import astropy.units as u
import numpy as np
from astropy.coordinates import CartesianRepresentation, EarthLocation
from astropy.time import Time, TimeDelta

from skywatch.access import Access
from skywatch.access.constraints import AzElRange, LineOfSight, Temporal
from skywatch.skypath import SkyPath
from skywatch.utils.coverage import GeoFilter, calculate_coverage
from skywatch.utils.funcs import lat_lon_to_xyz, FibonacciPointGenerator

from .utils import get_ephem_as_skypath


class SmokeTests(unittest.TestCase):

    def test_access(self):
        """
        Calculate very specific access from a ground station to a satellite.
        """

        t_start = Time("2024-02-01T00:00:00")
        t_end = Time("2024-02-02T00:00:00")
        low_fidelity_times = np.linspace(t_start, t_end, 1440)
        high_fidelity_times = np.linspace(t_start, t_end, 86400)

        earth_pos = SkyPath.from_body(low_fidelity_times, "earth")
        sun_pos = SkyPath.from_body(low_fidelity_times, "sun")
        moon_pos = SkyPath.from_body(low_fidelity_times, "moon")

        ground_station_pos = SkyPath.from_geodetic(low_fidelity_times[0], 34.5 * u.deg, -77.0 * u.deg, 0 * u.m)

        sat_position = get_ephem_as_skypath()

        t0 = time.time()
        access_times = (
            Access(
                LineOfSight(ground_station_pos, sat_position, earth_pos),
                AzElRange(ground_station_pos, sun_pos, min_el=0 * u.deg, max_el=5 * u.deg),
                LineOfSight(sat_position, sun_pos, earth_pos),
                Temporal(
                    min_time=Time("2024-02-01T22:15:00"),
                    max_time=Time("2024-02-01T22:17:00"),
                ),
                LineOfSight(moon_pos, sun_pos, earth_pos),
            )
            .use_precise_endpoints(True)
            .set_precision(0.001 * u.s)
            .set_min_duration(60 * u.s)
            .only_check_failed_constraints(True)
            .calculate_at(high_fidelity_times)
        )
        t1 = time.time()

        print(f"Access calculation took: {t1-t0} seconds")
        print(access_times.total_duration)
        print(access_times)

    def test_eclipse_access(self):
        """
        calculate approximate lunar and solar eclipse (total & partial) times until 2025
        """

        t_start = Time("2024-01-01T00:00:00")
        t_end = Time("2025-01-01T00:00:00")
        dt = TimeDelta(datetime.timedelta(hours=1), format="datetime")
        num_steps = int((t_end - t_start) / dt)
        times = np.linspace(t_start, t_end, num_steps)

        # define a sample of position around the surface of the moon
        moon_position_offsets = [lat_lon_to_xyz(i[0], i[1], radius=1079.6) for i in FibonacciPointGenerator(100)()]

        # get the moons position and add the offsets to the position to define
        # coordinates on the lunar surface
        moon_pos = SkyPath.from_body(times, "moon")
        moon_surface_samples = [
            SkyPath(
                CartesianRepresentation(*(moon_pos.cartesian.xyz.T + (moon_pos_offset * u.m)).T),
                frame="gcrs",
                obstime=times,
            )
            for moon_pos_offset in moon_position_offsets
        ]

        earth_pos = SkyPath.from_body(times, "earth")
        sun_pos = SkyPath.from_body(times, "sun")

        # calculate line of sight access from the sun to the moon when
        # obstructed by the earth
        t0 = time.time()
        lunar_access_interval = None
        for moon_sample in moon_surface_samples:
            lunar_eclipse_times = (
                Access(
                    LineOfSight(
                        sun_pos,
                        moon_sample,
                        earth_pos,
                        when_obstructed=True,
                        use_frame="gcrs",
                    )
                )
                .use_precise_endpoints(True)
                .set_precision(1 * u.s)
                .calculate_at(times)
            )

            # union the intervals together to get cumulative time
            if lunar_access_interval is None:
                lunar_access_interval = lunar_eclipse_times
            else:
                lunar_access_interval = lunar_access_interval | lunar_eclipse_times
        t1 = time.time()

        print(f"Lunar eclipse access calculation took: {t1-t0} seconds")
        print(f"There will be ~{lunar_access_interval.total_duration} seconds of lunar eclipse in 2024.")
        print(lunar_access_interval)
        self.assertTrue(len(lunar_access_interval) >= 1)

        # define a sample of points around the earth
        earth_positions = [
            SkyPath(EarthLocation(i[1] * u.deg, i[0] * u.deg, height=0 * u.m).get_itrs(t_start))
            for i in FibonacciPointGenerator(200)()
        ]

        # calculate line of sight access from points around the earth to the sun
        # when obstructed by the moon.
        solar_access_interval = None
        t0 = time.time()
        for earth_point in earth_positions:
            solar_eclipse_times = (
                Access(
                    LineOfSight(
                        earth_point,
                        sun_pos,
                        moon_pos,
                        sma=1079.6 * u.km,  # moon radius
                        smi=1079.6 * u.km,  # moon radius
                        when_obstructed=True,
                        use_frame="gcrs",
                    )
                )
                .use_precise_endpoints(True)
                .set_precision(1 * u.s)
                .calculate_at(times)
            )

            # union the intervals together to get cumulative time
            if solar_access_interval is None:
                solar_access_interval = solar_eclipse_times
            else:
                solar_access_interval = solar_access_interval | solar_eclipse_times
        t1 = time.time()

        print(f"Solar eclipse access calculation took: {t1-t0} seconds")
        print(f"There will be ~{solar_access_interval.total_duration} seconds of solar eclipse in 2024.")
        print(solar_access_interval)
        self.assertTrue(len(solar_access_interval) >= 2)

    def test_coverage(self):
        t_start = Time("2024-02-01T00:00:00")
        t_end = Time("2024-02-02T00:00:00")
        times = np.linspace(t_start, t_end, 8640)

        sat_position = get_ephem_as_skypath()

        worldwide_coverage_result = calculate_coverage([sat_position], times)
        print(f"Worldwide min revisit time: {worldwide_coverage_result.min_revisit_time}")
        print(f"Worldwide max revisit time: {worldwide_coverage_result.max_revisit_time}")

        at_equator_coverage = worldwide_coverage_result.filter(
            GeoFilter(min_latitude=-1 * u.deg, max_latitude=1 * u.deg)
        )
        print(f"Equatorial min revisit time: {at_equator_coverage.min_revisit_time}")
        print(f"Equatorial max revisit time: {at_equator_coverage.max_revisit_time}")



from abc import ABC, abstractmethod

from astropy.time import Time

from skywatch.look_angles.aert import AzElRangeTime
from skywatch.skypath import SkyPath


class BaseLookAngleStrategy(ABC):
    @abstractmethod
    def get_look_angles(self, target: SkyPath, time: Time) -> AzElRangeTime:
        """
        Get the Azimuth, Elevation, Range, and Time to a target at the given time(s).

        Args:
            target (SkyPath): SkyPath object to calculate look angles to.
            time (Time): Time(s) to calculate look angles to the target during.

        Returns:
            AzElRangeTime: Look angles to the target during the provided times.
        """
        pass



from typing import Union

import astropy.units as u
import numpy as np
from astropy.coordinates import BaseCoordinateFrame
from astropy.time import Time

from skywatch.access.base_constraint import BaseAccessConstraint
from skywatch.skypath import SkyPath


class LineOfSight(BaseAccessConstraint):
    def __init__(
        self,
        observer: SkyPath,
        target: SkyPath,
        obstructor: Union[SkyPath, BaseCoordinateFrame],
        sma=6378.137 * u.km,
        smi=6356.7523 * u.km,
        use_frame: str = "itrs",
        when_obstructed: bool = False,
    ) -> None:
        """
        Line of sight constraint which checks that the line of sight between two positions
        is not blocked by some body with a given semi major and semi minor access.

        Args:
            obstructor (Union[SkyPath, BaseCoordinateFrame]): coordinate of the body which potentially blocks line of sight.
            sma (Quantity, optional): semi major axis of the body. Defaults to 6378.137*u.km.
            smi (Quantity, optional): semi minor axis of the body. Defaults to 6356.7523*u.km.
            use_frame (str, optional): coordinate frame name to calculate access in. Defaults to 'itrs'.
        """
        super().__init__()
        assert observer is not target, "Observer cannot be the same object as the target"
        self.observer = observer
        self.target = target
        self.when_obstructed = when_obstructed

        self.sma = sma
        self.smi = smi
        self.use_frame = use_frame
        if isinstance(obstructor, SkyPath):
            self.obstructor = obstructor
        elif isinstance(obstructor, BaseCoordinateFrame):
            self.obstructor = SkyPath(obstructor)
        else:
            raise TypeError("Body must be a SkyPath, CoordinateInterpolator, or BaseCoordinateFrame.")

    def __call__(self, time: Time) -> np.ndarray:
        """
        Returns the subset of times when the earth does not obstruct the two Kinematics.

        Args:
            observer (CoordinateInterpolator): Coordinate to get line of sight access from
            target (CoordinateInterpolator): Coordinate to get line of sight access to
            time (Time): Times to check for line of sight access
            bounds_check (bool, optional): Whether or not to check the provided times are within the bounds
            of the observer and target CoordinateInterpolator.

        Returns:
            Time: boolean array of times when this SkyPath has line of sight access to the target.
        """
        pos1 = self.observer.state_at(time, self.use_frame, bounds_check=False).cartesian.xyz
        pos2 = self.target.state_at(time, self.use_frame, bounds_check=False).cartesian.xyz
        obstructor_pos = self.obstructor.state_at(time, self.use_frame, bounds_check=False).cartesian.xyz

        los_times = self._line_of_sight_body(pos1.T, pos2.T, obstructor_pos.T, self.sma, self.smi)
        if self.when_obstructed:
            return los_times
        return ~los_times

    @staticmethod
    def _line_of_sight_body(
        pos1: np.ndarray,
        pos2: np.ndarray,
        pos_body: np.ndarray,
        sma: u.km,
        smi: u.km,
    ) -> np.ndarray:
        """
        Determines whether a planetary body intersects the line of sight between two vectors.

        Parameters:
        pos1 (np.ndarray): The position vector of the observer in 3D space.
        pos2 (np.ndarray): The position vector of the target in 3D space.
        pos_body (np.ndarray): The position vector of the body (e.g., Earth) in 3D space.
        sma (u.km): The semi-major axis of the planetary body, default is 6378.137 km (Earth).
        smi (u.km): The semi-minor axis of the planetary body, default is 6356.752 km (Earth).

        Returns:
        np.ndarray: A boolean array indicating True when line of sight is not impeded by the body, and False when it is for each pair of position vectors.
        """
        # Make sure the arrays have the same shape
        assert pos1.shape == pos2.shape == pos_body.shape
        if pos1.ndim == 1:
            pos1 = np.reshape(pos1, (-1, 3))
            pos2 = np.reshape(pos2, (-1, 3))
            pos_body = np.reshape(pos_body, (-1, 3))

        # Calculate the line of sight vectors
        los_vectors = pos2 - pos1

        # Calculate the unit vectors of the line of sight
        los_unit_vectors = los_vectors / np.linalg.norm(los_vectors, axis=1)[:, np.newaxis]

        # Calculate the vectors from the observer to the body's center
        observer_to_body = pos_body - pos1

        # Calculate the scalar projections of observer_to_body onto the line of sight
        scalar_projections = np.einsum("ij,ij->i", observer_to_body, los_unit_vectors)

        # Calculate the closest points to the body's center on the line of sight
        closest_points = pos1 + scalar_projections[:, np.newaxis] * los_unit_vectors

        # Calculate the geocentric latitude of the closest points
        latitudes = np.arctan2(
            closest_points[:, 2],
            np.sqrt(closest_points[:, 0] ** 2 + closest_points[:, 1] ** 2),
        )

        # Calculate the body's radius at the geocentric latitude
        body_radius = np.sqrt(
            ((sma**2 * np.cos(latitudes)) ** 2 + (smi**2 * np.sin(latitudes)) ** 2)
            / ((sma * np.cos(latitudes)) ** 2 + (smi * np.sin(latitudes)) ** 2)
        )

        # Calculate the distances from the closest points to the body's center
        distances_to_body_center = np.linalg.norm(closest_points - pos_body, axis=1)

        # Check if the body intersects the line of sight
        intersections = np.logical_and(
            distances_to_body_center <= body_radius,
            np.logical_and(
                0 <= scalar_projections,
                scalar_projections <= np.linalg.norm(los_vectors, axis=1),
            ),
        )

        # Return the times that the body intersects the line of sight vector
        return intersections



import astropy.units as u
from astropy.coordinates import BaseCoordinateFrame, EarthLocation, SkyCoord, get_body
from astropy.time import Time


class SkyPathCreationMixin:

    @classmethod
    @u.quantity_input(latitude=u.deg, longitude=u.deg, altitude=u.m)
    def from_geodetic(
        cls,
        time: Time,
        latitude: u.Quantity,
        longitude: u.Quantity,
        altitude: u.Quantity,
    ):
        """
        Construct a SkyPath from a geodetic location.

        Args:
            time (Time): Time of the observation. For stationary coordinates, just use one time.
            latitude (u.Quantity): latitude in degrees.
            longitude (u.Quantity): longitude in degrees.
            altitude (u.Quantity): altitude in meters.

        Returns:
            SkyPath: SkyPath instance in the ITRS frame.
        """
        return cls(
            SkyCoord(
                EarthLocation(lat=latitude, lon=longitude, height=altitude).get_itrs(
                    obstime=time
                )
            )
        )

    @classmethod
    @u.quantity_input(x=u.m, y=u.m, z=u.m, v_x=u.m / u.s, v_y=u.m / u.s, v_z=u.m / u.s)
    def from_ECEF(
        cls,
        time: Time,
        x: u.Quantity,
        y: u.Quantity,
        z: u.Quantity,
        v_x: u.Quantity = None,
        v_y: u.Quantity = None,
        v_z: u.Quantity = None,
    ):
        """
        Construct a SkyPath from ECEF X,Y,Z position and/or
        V_X,V_Y,V_Z velocity at a given time.

        NOTE: ECEF is NOT equivalent to the ITRS frame; however, it is
        often considered close enough and functionally comparable.

        Args:
            time (Time): Time of position and/or velocity measurements.
            x (u.Quantity): position X
            y (u.Quantity): position Y
            z (u.Quantity): posizion Z
            v_x (u.Quantity, optional): velocity X. Defaults to None.
            v_y (u.Quantity, optional): velocity Y. Defaults to None.
            v_z (u.Quantity, optional): velocity Z. Defaults to None.

        Returns:
            SkyPath: SkyPath instance in the ITRS frame.
        """
        return cls(
            SkyCoord(
                x=x,
                y=y,
                z=z,
                v_x=v_x,
                v_y=v_y,
                v_z=v_z,
                frame="itrs",
                representation_type="cartesian",
                obstime=time,
            )
        )

    @classmethod
    @u.quantity_input(x=u.m, y=u.m, z=u.m, v_x=u.m / u.s, v_y=u.m / u.s, v_z=u.m / u.s)
    def from_ECI(
        cls,
        time: Time,
        x: u.Quantity,
        y: u.Quantity,
        z: u.Quantity,
        v_x: u.Quantity = None,
        v_y: u.Quantity = None,
        v_z: u.Quantity = None,
    ):
        """
        Construct a SkyPath from ECI X,Y,Z position and/or
        V_X,V_Y,V_Z velocity at a given time.

        Args:
            time (Time): Time of position and/or velocity measurements.
            x (u.Quantity): position X
            y (u.Quantity): position Y
            z (u.Quantity): posizion Z
            v_x (u.Quantity, optional): velocity X. Defaults to None.
            v_y (u.Quantity, optional): velocity Y. Defaults to None.
            v_z (u.Quantity, optional): velocity Z. Defaults to None.

        Returns:
            SkyPath: SkyPath instance in the GCRS frame.
        """
        return cls(
            SkyCoord(
                x=x,
                y=y,
                z=z,
                v_x=v_x,
                v_y=v_y,
                v_z=v_z,
                frame="gcrs",
                representation_type="cartesian",
                obstime=time,
            )
        )

    @classmethod
    def from_body(
        cls,
        time: Time,
        body: str,
        location: EarthLocation = None,
        ephemeris: str = None,
    ):
        """
        Construct a SkyPath for a solar system body as observed
        from a location on Earth in the `~astropy.coordinates.GCRS` reference
        system.

        Parameters
        ----------
        body : str or list of tuple
            The solar system body for which to calculate positions.  Can also be a
            kernel specifier (list of 2-tuples) if the ``ephemeris`` is a JPL
            kernel.
        time : `~astropy.time.Time`
            Time of observation.
        location : `~astropy.coordinates.EarthLocation`, optional
            Location of observer on the Earth.  If not given, will be taken from
            ``time`` (if not present, a geocentric observer will be assumed).
        ephemeris : str, optional
            Ephemeris to use.  If not given, use the one set with
            ``astropy.coordinates.solar_system_ephemeris.set`` (which is
            set to 'builtin' by default).

        Returns:
            SkyPath: SkyPath instance representing the coordinates of a planetary body.
        """
        return cls(get_body(body, time, location, ephemeris))

    @classmethod
    def from_base_coordinate(cls, base: BaseCoordinateFrame):
        """
        Construct a SkyPath from a base coordinate frame like ITRS, GCRS, ICRS, etc.

        Args:
            base (BaseCoordinateFrame): base coordinate frame.

        Returns:
            SkyPath: SkyPath instance wrapping the base coordinate frame.
        """
        return cls(base)



import astropy.units as u
import numpy as np
import pymap3d
from astropy.time import Time

from skywatch.access.base_constraint import BaseAccessConstraint
from skywatch.skypath import SkyPath
from pymap3d import ecef2geodetic, ecef2aer


class AzElRange(BaseAccessConstraint):
    @u.quantity_input(
        min_az=u.deg,
        max_az=u.deg,
        min_el=u.deg,
        max_el=u.deg,
        min_range=u.m,
        max_range=u.m,
    )
    def __init__(
        self,
        observer: SkyPath,
        target: SkyPath,
        min_az: u.deg = 0 * u.deg,
        max_az: u.deg = 360 * u.deg,
        min_el: u.deg = 0 * u.deg,
        max_el: u.deg = 90 * u.deg,
        min_range: u.m = 0 * u.m,
        max_range: u.m = np.inf * u.m,
    ) -> None:
        """
        Determine access using ENU azimuth, elevation, and range values from an earth based position.

        Args:
            min_az (u.deg, optional): minimum allowable azimuth angle to target. Defaults to 0*u.deg.
            max_az (u.deg, optional): maximum allowable azimuth angle to target. Defaults to 360*u.deg.
            min_el (u.deg, optional): minimum allowable elevation angle to target. Defaults to 0*u.deg.
            max_el (u.deg, optional): maximum allowable elevation angle to target. Defaults to 90*u.deg.
            min_range (u.m, optional): minimum allowable range angle to target. Defaults to 0*u.m.
            max_range (u.m, optional): maximum allowable range to target. Defaults to np.inf*u.m.
        """
        super().__init__()
        self.observer = observer
        self.target = target
        self.min_az = min_az
        self.max_az = max_az
        self.min_el = min_el
        self.max_el = max_el
        self.min_range = min_range
        self.max_range = max_range

    def __call__(self, time: Time) -> np.ndarray:
        """Constraints the times when the azimuth, elevation, and range values are satisfied from the observer to the target.

        Args:
            time (Time): time to check for azimuth, elevation and range constraints

        Returns:
            np.ndarray: boolean array representing times when this constraint succeeds vs. fails.
        """

        # NOTE: Using pymap3d is tremendously faster than having astropy calculate the AltAz frame from the observer to the target,
        # however there are cases for distant objects that this loses accuracy.
        az, el, rng = ecef2aer(
            *self.target.state_at(time, "itrs", bounds_check=False).cartesian.xyz.to_value(u.m),
            *ecef2geodetic(*self.observer.state_at(time, "itrs", bounds_check=False).cartesian.xyz.to_value(u.m))
        )
        results = (
            (az >= self.min_az.value)
            & (az <= self.max_az.value)
            & (el >= self.min_el.value)
            & (el <= self.max_el.value)
            & (rng >= self.min_range.value)
            & (rng <= self.max_range.value)
        )
        return results



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import List

import astropy.units as u
import numpy as np
import tqdm
from astropy.time import Time

from skywatch.access import Access
from skywatch.access.constraints import AzElRange
from skywatch.skypath import SkyPath
from skywatch.time.interval import Interval
from skywatch.utils.funcs import PointGenerator, FibonacciPointGenerator

import matplotlib.pyplot as plt
import pyvista as pv
from mpl_toolkits.basemap import Basemap
from pyvista import examples
from scipy.interpolate import griddata
from skywatch.skypath import SkyPath


@dataclass
class SatelliteAccess:
    satellite: SkyPath
    interval: Interval


class CoveragePoint:
    def __init__(
        self,
        latitude: u.deg,
        longitude: u.deg,
        time_bound_lower: Time,
        time_bound_upper: Time,
    ) -> None:
        self.latitude = latitude
        self.longitude = longitude
        self.time_bound_lower = time_bound_lower
        self.time_bound_upper = time_bound_upper

        self.interval = Interval()
        self.satellite_accesses = list()

    def _default_revisit_stats(self) -> u.s:
        if self.interval.lower == self.time_bound_lower and self.interval.upper == self.time_bound_upper:
            return 0 * u.s
        return np.inf * u.s

    @property
    def satellite_visibility_count(self) -> int:
        return len([i for i in self.satellite_accesses if len(i.interval._intervals) != 0])

    @property
    def max_simulatenous_visibility(self) -> int:
        # Create a list of all start and end times
        events = []
        for _access in self.satellite_accesses:
            for _interval in _access.interval._intervals:
                events.append((_interval.lower, 1))  # Start of an interval
                events.append((_interval.upper, -1))  # End of an interval

        # Sort the events
        events.sort()

        max_simulatenous = 0
        current_simulatenous = 0

        # Sweep line algorithm
        for _, event_type in events:
            current_simulatenous += event_type
            max_simulatenous = max(max_simulatenous, current_simulatenous)

        return max_simulatenous

    @property
    def cumulative_access(self) -> u.s:
        cumulative_access = 0.0 * u.s
        for sat_access in self.satellite_accesses:
            cumulative_access += sat_access.interval.total_duration
        return cumulative_access

    @property
    def average_revisit_time(self) -> u.s:
        """
        Average revisit time is the time between the end of one access interval and the start of the next, all summed together.

        Returns:
            u.s: Returns np.inf * u.s if this coordinate has no access, returns 0 seconds if access
            was never lost, else returns seconds representing the average of all revisit times for this lat/lon coordinate.
        """
        if len(self.interval) < 2:
            return self._default_revisit_stats()

        lost_access_times = 0 * u.s
        for index, window in enumerate(self.interval[:-1]):
            lost_access_times += (self.interval[index + 1].lower - window.upper).datetime.total_seconds() * u.s

        return lost_access_times / (len(self.interval) - 1)

    @property
    def max_revisit_time(self) -> u.s:
        """
        Maximum revisit time is the time between the end of one access interval and the start of the next.

        Returns:
            u.s: Returns None if this coordinate has no access, returns 0 seconds if access
            was never lost, else returns seconds representing the average of all revisit times for this lat/lon coordinate.
        """
        if len(self.interval) < 2:
            return self._default_revisit_stats()

        max_revisit_time = None
        for index, window in enumerate(self.interval[:-1]):
            loss_seconds = (self.interval[index + 1].lower - window.upper).datetime.total_seconds() * u.s
            if max_revisit_time == None:
                max_revisit_time = loss_seconds
            elif max_revisit_time < loss_seconds:
                max_revisit_time = loss_seconds

        return max_revisit_time

    @property
    def min_revisit_time(self) -> u.s:
        """
        Maximum revisit time is the time between the end of one access interval and the start of the next.

        Returns:
            u.s: Returns None if this coordinate has no access, returns 0 seconds if access
            was never lost, else returns seconds representing the average of all revisit times for this lat/lon coordinate.
        """
        if len(self.interval) < 2:
            return self._default_revisit_stats()

        min_revisit_time = None
        for index, window in enumerate(self.interval[:-1]):
            loss_seconds = (self.interval[index + 1].lower - window.upper).datetime.total_seconds() * u.s
            if min_revisit_time == None:
                min_revisit_time = loss_seconds
            elif min_revisit_time > loss_seconds:
                min_revisit_time = loss_seconds

        return min_revisit_time


class CoverageFilter(ABC):
    @abstractmethod
    def __call__(self, coverage_point: CoveragePoint) -> bool:
        """Return True when filter should remove the CoveragePoint, false when it should remain.

        Args:
            coverage_point (CoveragePoint): Point to check if it satisfies the filter.

        Returns:
            bool: True or False representing filter result.
        """
        pass


class GeoFilter(CoverageFilter):
    def __init__(
        self,
        min_latitude: u.deg = -90 * u.deg,
        max_latitude: u.deg = 90 * u.deg,
        min_longitude: u.deg = -180 * u.deg,
        max_longitude: u.deg = 180 * u.deg,
    ) -> None:
        super().__init__()
        self.min_latitude = min_latitude
        self.max_latitude = max_latitude
        self.min_longitude = min_longitude
        self.max_longitude = max_longitude

    def __call__(self, coverage_point: CoveragePoint) -> bool:
        if coverage_point.latitude < self.min_latitude:
            return True
        if coverage_point.latitude > self.max_latitude:
            return True
        if coverage_point.longitude < self.min_longitude:
            return True
        if coverage_point.longitude > self.max_longitude:
            return True
        return False


class CoverageResult:
    def __init__(self, *coverage_points) -> None:
        # statistics parameters
        self.percent_coverage = 0.0
        self.max_duration = 0.0 * u.s
        self.min_duration = None
        self.average_duration = 0.0 * u.s
        self.total_coverage_duration = 0.0 * u.s
        self.num_points_with_coverage = 0
        self.num_points_total = 0

        self.coverage_points = []
        for result in coverage_points:
            self.add_result(result)

    def add_result(self, coverage: CoveragePoint):
        if not isinstance(coverage, CoveragePoint):
            raise TypeError("Coverage must be of type CoveragePoint")

        self.coverage_points.append(coverage)
        self.num_points_total += 1

        # calculate statistics
        coverage_duration = coverage.cumulative_access
        self.total_coverage_duration += coverage_duration

        if coverage_duration > self.max_duration:
            self.max_duration = coverage_duration

        if self.min_duration == None:
            self.min_duration = coverage_duration
        elif self.min_duration > coverage_duration:
            self.min_duration = coverage_duration

        if coverage_duration > 0.0 * u.s:
            self.num_points_with_coverage += 1
        self.percent_coverage = self.num_points_with_coverage / self.num_points_total

        self.average_duration = self.total_coverage_duration / self.num_points_total

    def get_max_simulatenous_visibility(self) -> int:
        maximum = 0
        for result in tqdm.tqdm(
            self.coverage_points,
            desc="Calculating maximum simulateous coverage:",
        ):
            num_ball = result.max_simulatenous_visibility
            if num_ball > maximum:
                maximum = num_ball
        return maximum

    @property
    def max_revisit_time(self) -> u.s:
        _inf = np.inf * u.s

        max_revisit = None
        for result in self.coverage_points:
            result: CoveragePoint
            result_max = result.max_revisit_time
            if result_max == _inf:
                continue

            if max_revisit == None:
                max_revisit = result_max
                continue

            if result_max > max_revisit:
                max_revisit = result_max

        return max_revisit

    @property
    def min_revisit_time(self) -> u.s:
        _inf = np.inf * u.s

        min_revisit = None
        for result in self.coverage_points:
            result: CoveragePoint
            result_min = result.min_revisit_time
            if result_min == _inf:
                continue

            if min_revisit == None:
                min_revisit = result_min
                continue

            if result_min < min_revisit:
                min_revisit = result_min

        return min_revisit

    def filter(self, *filters: CoverageFilter):
        # check filter types
        for filter in filters:
            if not isinstance(filter, CoverageFilter):
                raise TypeError("Filters must be of type CoverageFilter")

        # run the filters over the results
        new_coverage = CoverageResult()
        for result in self.coverage_points:
            keep = True
            for filter in filters:
                if filter(result):
                    keep = False
                    break
            if keep:
                new_coverage.add_result(result)

        # return remaining coverage results
        return new_coverage

    def __iter__(self):
        return iter(self.coverage_points)

    def __len__(self) -> int:
        return len(self.coverage_points)

    @staticmethod
    def _wrap_longitude_data(longitudes, latitudes, cumulative_access):
        """
        Wrap longitudes around the -180/180 boundary to ensure continuity in interpolation.
        """
        # Duplicate the data for continuity around the 180/-180 boundary
        wrapped_longitudes = np.concatenate([longitudes, longitudes + 360, longitudes - 360])
        wrapped_latitudes = np.concatenate([latitudes, latitudes, latitudes])
        wrapped_cumulative_access = np.concatenate([cumulative_access, cumulative_access, cumulative_access])

        return wrapped_longitudes, wrapped_latitudes, wrapped_cumulative_access

    def plot(
        self,
        bin_resolution: float = 0.1,
        map_resolution: str = "c",
        map_projection: str = "cyl",
        grid_interpolation_method: str = "linear",
        cmap_theme: str = "jet_r",
        show: bool = True,
    ) -> tuple:
        """
        Create a contour plot on the surface of the Earth, colored by cumulative access time
        at each latitude and longitude.

        Parameters:
        coverage_points (List[CoveragePoint]): List of CoveragePoint objects with cumulative access times.
        resolution (float): The resolution for the grid used in contouring.
        """
        coverage_points = list(self)

        # Extract latitude, longitude, and cumulative access from each CoveragePoint
        latitudes = np.array([cp.latitude.to_value(u.deg) for cp in coverage_points])
        longitudes = np.array([cp.longitude.to_value(u.deg) for cp in coverage_points])
        cumulative_access = np.array(
            [
                (cp.cumulative_access / (cp.time_bound_upper - cp.time_bound_lower).to(u.day)).to_value(u.hr / u.day)
                for cp in coverage_points
            ]
        )

        # Wrap the longitude data around the -180/180 boundary
        longitudes, latitudes, cumulative_access = self._wrap_longitude_data(longitudes, latitudes, cumulative_access)

        # Create a Basemap instance
        fig, ax = plt.subplots(figsize=(12, 8))
        m = Basemap(
            projection=map_projection,
            llcrnrlat=-90,
            urcrnrlat=90,
            llcrnrlon=-180,
            urcrnrlon=180,
            resolution=map_resolution,
            ax=ax,
        )

        # Define a grid for the contour plot
        lon_bins = np.arange(-180, 180 + bin_resolution, bin_resolution)
        lat_bins = np.arange(-90, 90 + bin_resolution, bin_resolution)
        lon_grid, lat_grid = np.meshgrid(lon_bins, lat_bins)

        # Interpolate cumulative access to the grid
        grid_cumulative_access = griddata(
            (longitudes, latitudes),
            cumulative_access,
            (lon_grid, lat_grid),
            method=grid_interpolation_method,
            fill_value=0,
        )

        # Generate the contour plot on the map
        x, y = m(lon_grid, lat_grid)
        contour = m.contourf(x, y, grid_cumulative_access, cmap=cmap_theme, levels=60, alpha=0.30)

        # Add map features
        m.drawparallels(np.arange(-90, 90, 30), labels=[1, 0, 0, 0])
        m.drawmeridians(np.arange(-180, 180, 60), labels=[0, 0, 0, 1])
        m.shadedrelief(scale=0.1)

        # Add color bar
        plt.colorbar(contour, label="Cumulative Access (Hours Per Day)", ax=ax)

        # Set title
        plt.title("Contour Plot of Cumulative Access on Earth's Surface")

        if show:
            plt.show()

        # Return the figure and axes objects
        return fig, ax

    def plot_3d(
        self,
        grid_interpolation_method: str = "linear",
        cmap_theme: str = "jet_r",
        add_background: bool = True,
        scale_factor: float = 1.01,
    ):
        """
        Create a 3D contour plot on the surface of the Earth, colored by cumulative access time,
        and overlay it on an Earth texture using pyvista.
        """
        coverage_points = list(self)

        # Extract latitude, longitude, and cumulative access from each CoveragePoint
        latitudes = np.array([cp.latitude.to(u.deg).value for cp in coverage_points])
        longitudes = np.array([cp.longitude.to(u.deg).value for cp in coverage_points])
        cumulative_access = np.array(
            [
                (cp.cumulative_access / (cp.time_bound_upper - cp.time_bound_lower).to(u.day)).to_value(u.hr / u.day)
                for cp in coverage_points
            ]
        )

        # Wrap the longitude data around the -180/180 boundary
        longitudes, latitudes, cumulative_access = self._wrap_longitude_data(longitudes, latitudes, cumulative_access)

        # Load the pre-textured Earth sphere from pyvista examples
        mesh = examples.planets.load_earth()
        texture = examples.load_globe_texture()

        # Create a copy of the mesh and scale it slightly larger for the cumulative access data
        cumulative_mesh = mesh.copy()
        cumulative_mesh.points *= scale_factor

        # Interpolate cumulative access data onto the cumulative_mesh's points
        # Convert cumulative_mesh points to latitude and longitude
        x, y, z = (
            cumulative_mesh.points[:, 0],
            cumulative_mesh.points[:, 1],
            cumulative_mesh.points[:, 2],
        )
        lat = np.degrees(np.arcsin(z / np.sqrt(x**2 + y**2 + z**2)))
        lon = np.degrees(np.arctan2(y, x))

        # Interpolate cumulative access onto the cumulative_mesh's points
        interpolated_access = griddata(
            (longitudes, latitudes),
            cumulative_access,
            (lon, lat),
            method=grid_interpolation_method,
            fill_value=0,
        )

        # Add the cumulative access data as scalars to the cumulative_mesh
        cumulative_mesh["cumulative_access"] = interpolated_access

        # Create a plotter with depth peeling to reduce rendering artifacts
        plotter = pv.Plotter()
        plotter.enable_depth_peeling()  # Enable depth peeling to fix transparency issues

        # Add a background image of stars
        if add_background:
            image_path = examples.planets.download_stars_sky_background(load=False)
            plotter.add_background_image(image_path)

        # Add the textured Earth mesh
        plotter.add_mesh(mesh, texture=texture)

        # Overlay the cumulative access data as a semi-transparent layer on the scaled-up mesh
        actor = plotter.add_mesh(
            cumulative_mesh,
            scalars="cumulative_access",
            cmap=cmap_theme,
            opacity=0.3,
            show_scalar_bar=False,
            lighting=True,
        )
        actor_prop = actor.GetProperty()

        plotter.add_scalar_bar(title="Cumulative Access (Hours/Day)", color="white", interactive=False)

        # Slider callback function to adjust opacity
        def set_opacity(value):
            actor_prop.opacity = value

        # Add a slider widget to control opacity
        plotter.add_slider_widget(
            callback=set_opacity,
            rng=[0.0, 1.0],
            value=0.3,
            title="Opacity",
            pointa=(0.1, 0.1),
            pointb=(0.25, 0.1),
            color="white",
            title_color="white",
            interaction_event="always",
        )

        # Remove axes for a cleaner look
        plotter.hide_axes()

        plotter.set_background("black")

        # Show the plot
        plotter.show()


@u.quantity_input
def calculate_coverage(
    satellites: List[SkyPath],
    time: Time,
    min_elevation: u.Quantity[u.deg] = 0 * u.deg,
    precision: u.Quantity[u.s] = 0.1 * u.s,
    use_precise_endpoints: bool = False,
    min_duration: u.Quantity[u.s] = 0.0 * u.s,
    point_generator: PointGenerator = FibonacciPointGenerator(1000),
) -> CoverageResult:

    earth_point_coverages = []
    earth_points = []
    for lat, lon in point_generator():
        lat, lon = lat * u.deg, lon * u.deg
        earth_points.append(SkyPath.from_geodetic(time[0], lat, lon, 0 * u.m))
        earth_point_coverages.append(CoveragePoint(lat, lon, time[0], time[-1]))

    for point, coverage in tqdm.tqdm(
        zip(earth_points, earth_point_coverages),
        desc="Calculating Coverage",
        total=len(earth_points),
    ):
        for satellite in satellites:
            _access = (
                Access(
                    AzElRange(point, satellite, min_el=min_elevation),
                )
                .use_precise_endpoints(use_precise_endpoints)
                .set_precision(precision)
                .set_min_duration(min_duration)
                .calculate_at(time)
            )
            coverage.interval = coverage.interval.union(_access)
            coverage.satellite_accesses.append(SatelliteAccess(satellite, _access))

    return CoverageResult(*earth_point_coverages)



from astropy.time import Time
from scipy.spatial.transform import Rotation

from skywatch.attitude.base_attitude import BaseAttitudeStrategy


class Fixed(BaseAttitudeStrategy):
    def __init__(self, rotation: Rotation) -> None:
        """
        Provide a fixed single attitude as the attitude for all time durations.

        Args:
            rotation (Rotation): Rotation to return at all time steps.
        """
        self.rotation = rotation

    def at(self, time: Time) -> Rotation:
        return self.rotation



import time

import numpy as np
from astropy.time import Time
from utils import get_ephem_as_skypath

from skywatch.skypath import SkyPath
from skywatch.utils import coverage

if __name__ == "__main__":
    t_start = Time("2024-02-01T00:00:00")
    t_end = Time("2024-02-02T00:00:00")
    times = np.linspace(t_start, t_end, 8640)
    sat_path = get_ephem_as_skypath()

    reversed_sat_path = SkyPath(
        sat_path.cartesian[::-1],
        representation_type="cartesian",
        frame="gcrs",
        obstime=sat_path.obstime,
    )

    t0 = time.time()
    coverage_result = coverage.calculate_coverage(
        [sat_path, reversed_sat_path] * 1,
        times,
        use_precise_endpoints=False,
    )
    t1 = time.time()
    print(f"Coverage calculation took: {t1-t0} seconds")

    fig, ax = coverage_result.plot()
    coverage_result.plot_3d()

    t0 = time.time()
    max_num_ball = coverage_result.get_max_simulatenous_visibility()
    print(f"Max Num Ball: {max_num_ball}")
    t1 = time.time()
    print(f"Max num ball took: {t1-t0} seconds")

    # t0 = time.time()
    # data, points = [], []
    # for point in coverage_result:
    #     xyz = SkyPath.from_geodetic(times[0], point.latitude, point.longitude, 0 * u.m).cartesian.xyz.to(u.km).value
    #     xyz[2] += 0.01
    #     points.append(xyz / (np.linalg.norm(xyz)))
    #     data.append(point.cumulative_access.value)

    # # Create a GeoPlotter object
    # plotter = gv.GeoPlotter()
    # m = pv.PolyData(points).delaunay_3d()
    # m["data"] = data
    # plotter.add_mesh(m, smooth_shading=True, scalars=data)

    # plotter.add_coastlines(zlevel=1, color="black")

    # t1 = time.time()
    # print(f"Geovista plotting took: {t1-t0} seconds")

    # # Display the plot
    # plotter.show()



import astropy.units as u
import numpy as np
from astropy.coordinates import SkyCoord
from astropy.time import Time
from scipy.interpolate import CubicHermiteSpline, CubicSpline

from .creation import SkyPathCreationMixin


class SkyPath(SkyCoord, SkyPathCreationMixin):

    class SavedFrame:
        """
        Represents the a coordinate frames interpolation splines for quick position and velocity calculations.
        """

        def __init__(
            self,
            name: str,
            position_spline: CubicHermiteSpline,
            velocity_spline: CubicHermiteSpline,
        ) -> None:
            self.name = name
            self.position_spline = position_spline
            self.velocity_spline = velocity_spline

        def __repr__(self) -> str:
            return self.name

    def _copy_from(self, other: "SkyPath"):
        self._interpolation_allowed = other._interpolation_allowed
        self._min_original_time = other._min_original_time
        self._max_original_time = other._max_original_time
        self._saved_frames = other._saved_frames
        self._original_times = other._original_times
        self._original_frame = other._original_frame

    def __init__(self, *args, **kwargs) -> None:
        """
        SkyPath has equivalent functionality to the Astropy SkyCoord class, except
        it includes interpolation to get the coordinate state(s) at any time
        between the minimum and maximum obstime of the coordinates that this instance
        was created with.

        This class introduces a new *state_at* method that acts as the
        SkyCoord.transform_to(*frame*) method, but saves the state of this SkyCoord
        in the new frame for future calls to the state_at method. Future calls return
        the position of the SkyPath in the given frame using the saved and interpolated
        state. This allows position and velocity to be determined at any point in time
        for a SkyPath, even if the time was not present in the obstime array upon creation
        of the object.

        Therefore, this class is useful for simulations where you may want the coordinates
        of an object in multiple frames at multiple points in time, but dont want to pay the
        cost to transform the object into the new frame every time you need the coordinates.
        """
        super().__init__(*args, **kwargs)

        # call copy constructor
        if len(args) != 0 and isinstance(args[0], SkyPath):
            self._copy_from(args[0])

        # construct new instance
        else:
            # check that the obstime has more than 1 value in it for interpolation
            if self.data.size > 1:
                self._interpolation_allowed = True
            else:
                self._interpolation_allowed = False

            self._min_original_time = None
            self._max_original_time = None
            self._saved_frames = []
            self._original_times = self.obstime
            self._original_frame = self

    @property
    def min_time(self) -> Time:
        if self._min_original_time is not None:
            return self._min_original_time
        self._min_original_time = np.min(self._original_times)
        return self._min_original_time

    @property
    def max_time(self) -> Time:
        if self._max_original_time is not None:
            return self._max_original_time
        self._max_original_time = np.max(self._original_times)
        return self._max_original_time

    def state_at(self, time: Time, frame: str, copy: bool = True, bounds_check: bool = True) -> "SkyPath":
        """
        Interpolates this SkyPath at the given time(s) and saves the
        interpolation spline for later use. Therefore, you only pay the computation penalty
        for frame transforms once, and every subsequent call this method uses the saved frames
        interpolators to provide quick coordinate results for the provided time(s).

        Args:
            time (Time): time(s) to get the coordinate frame values for.
            frame (str): BaseCoordinateFrame name to get the results in.
            Example: "itrs", "gcrs", "icrs", "teme", etc...

        Returns:
            SkyPath: SkyPath representing an Astropy SkyCoord in the coordinate system you requested.
        """
        if not self._interpolation_allowed:
            return SkyPath(SkyCoord(self.frame.cartesian, obstime=time, frame=self.frame.name).transform_to(frame))

        if bounds_check:
            if np.min(time) < self.min_time or np.max(time) > self.max_time:
                raise ValueError(
                    f"Cannot interpolate times that are outside the bounds of the original coordinate frame.\nTime bounds are: [{self._min_original_time}, {self._max_original_time}]"
                )

        # using a list for speed, check if the conversion has already been calculated
        saved_frame = None
        for f in self._saved_frames:
            if f.name == frame:
                saved_frame = f
                break

        # if the frame exists, use it to interpolate the position and velocity to the given time
        unix_time = time.unix
        if saved_frame is not None:
            _position = saved_frame.position_spline(unix_time) * u.m
            if saved_frame.velocity_spline is not None:
                _velocity = saved_frame.velocity_spline(unix_time) * (u.m / u.s)
            else:
                _velocity = [None, None, None]

            # return a new copy of this object in the requested frame
            new_coord = SkyPath(
                x=_position[0],
                y=_position[1],
                z=_position[2],
                v_x=_velocity[0],
                v_y=_velocity[1],
                v_z=_velocity[2],
                frame=saved_frame.name,
                representation_type="cartesian",
                obstime=time,
                copy=False,
            )

            if copy:
                new_coord._copy_from(self)

            return new_coord

        # we have not converted to this frame before, so we need to do the conversion and store it as a SavedFrame
        try:
            converted_frame = self._original_frame.transform_to(frame)
        except AttributeError as err:
            # handle errors when coming from non terrestrial coordinate system
            if err.name == "to_geodetic":
                converted_frame = self.state_at(self._original_times, "itrs", copy=False).transform_to(frame)
            else:
                raise err

        # get the position and velocity in cartesian XYZ coordinates of the converted frame
        position = converted_frame.cartesian.xyz.to(u.m)
        differentials = converted_frame.cartesian.differentials.get("s", None)
        if differentials is not None:
            velocity = differentials.d_xyz.to(u.m / u.s)
        else:
            velocity = None

        # now create the splines for the data that we have
        if velocity is not None:
            position_spline = CubicHermiteSpline(
                self._original_times.unix, position, velocity, axis=1, extrapolate=False
            )
            velocity_spline = position_spline.derivative()
            interpolated_position = position_spline(unix_time) * u.m
            interpolated_velocity = velocity_spline(unix_time) * (u.m / u.s)
        else:
            position_spline = CubicSpline(self._original_times.unix, position, axis=1, extrapolate=False)
            velocity_spline = None
            interpolated_position = position_spline(unix_time) * u.m
            interpolated_velocity = [None, None, None]

        # save the frame and its splines
        new_frame = self.SavedFrame(frame, position_spline, velocity_spline)
        self._saved_frames.append(new_frame)

        # construct the new frame from the interpolated coordinates as a copy of this object
        new_coord = SkyPath(
            x=interpolated_position[0],
            y=interpolated_position[1],
            z=interpolated_position[2],
            v_x=interpolated_velocity[0],
            v_y=interpolated_velocity[1],
            v_z=interpolated_velocity[2],
            frame=frame,
            representation_type="cartesian",
            obstime=time,
            copy=False,
        )

        if copy:
            new_coord._copy_from(self)

        return new_coord



import numpy as np
from astropy.time import Time

from skywatch.access.base_constraint import BaseAccessConstraint


class Temporal(BaseAccessConstraint):
    def __init__(self, min_time: Time = None, max_time: Time = None, inner: bool = True) -> None:
        """
        Only allows times in the time array that are within self.from_time and self.to_time (if self.inner is True)
        to pass the constraint.

        If self.inner is False, only times that do not fall between the self.from_time and self.to_time are allowed
        to pass this constraint.

        Args:
            from_time (Time): Lower time bound.
            to_time (Time): Upper time bound.
            inner (bool, optional): If True, only times within the lower time bound and upper time bound will pass
            this constraint. If False, only times outside the time bounds will pass. Defaults to True.
        """
        super().__init__()
        self.min_time = min_time
        self.max_time = max_time
        self.inner = inner

        if self.min_time is None and self.max_time is None:
            raise ValueError("Min time and Max time cannot both be None. You must set one, or both.")

    def __call__(self, time: Time) -> np.ndarray:
        """
        Compute times that pass this constraint.

        Args:
            observer (CoordinateInterpolator): unused, can be None.
            target (CoordinateInterpolator): unused, can be None.
            time (Time): times to check against this constraints allowed times.
            bounds_check (bool, optional): unused. Defaults to True.

        Returns:
            np.ndarray: Boolean array representing times that pass this constraint.
        """
        if self.min_time is not None and self.max_time is not None:
            if self.inner:
                return (time >= self.min_time) & (time <= self.max_time)
            else:
                return (time <= self.min_time) | (time >= self.max_time)

        if self.min_time is not None:
            return time >= self.min_time

        if self.max_time is not None:
            return time <= self.max_time



import astropy.units as u
import portion as P
from astropy.time import Time


class Interval(P.Interval):
    def __init__(self, *intervals):
        """
        Represents an interval of time.
        """
        super().__init__(*intervals)

    @property
    def total_duration(self) -> u.s:
        """
        Duration of all sub-intervals added together.

        Returns:
            u.s: total duration in seconds
        """
        duration = 0
        for interval in self:
            duration += (
                Time(interval.upper) - Time(interval.lower)
            ).datetime.total_seconds()
        return duration * u.s



import astropy.units as u
import numpy as np
from astropy.time import Time
from scipy.spatial.transform import Rotation

from skywatch.attitude.base_attitude import BaseAttitudeStrategy
from skywatch.skypath.skypath import SkyPath


class LVLH(BaseAttitudeStrategy):
    def __init__(
        self, observer: SkyPath, frame: str = "gcrs", offset: Rotation = None
    ) -> None:
        """
        Defines the LVLH orbital frame using X as the direction of forward velocity,
        Z as the direction to the center of the reference frame, and Y as the cross
        product completing the right-handed system.

        Args:
            observer (SkyPath): Observer coordinates
            frame (str, optional): Reference frame for the LVLH calculations. Defaults to "gcrs".
            offset (Rotation, optional): Additional rotation to apply to the LVLH frame. Defaults to None.
        """
        self.frame = frame
        self.observer = observer
        self.offset = offset

    def at(self, time: Time) -> Rotation:
        state = self.observer.state_at(time, frame=self.frame)
        pos = state.cartesian.xyz.to(u.m).value
        differentials = state.cartesian.differentials.get("s", None)
        if differentials is None:
            raise AttributeError(
                "Observer has no velocity component. Velocity is needed to calculate the LVLH frame."
            )
        vel = differentials.d_xyz.to(u.m / u.s).value

        # Now you have the
        rot_matrix = Rotation.from_matrix(LVLH.calculate_reference_frame(*pos, *vel))
        if self.offset is not None:
            rot_matrix = rot_matrix * self.offset
        return rot_matrix

    @staticmethod
    def calculate_reference_frame(
        x: np.ndarray,
        y: np.ndarray,
        z: np.ndarray,
        v_x: np.ndarray,
        v_y: np.ndarray,
        v_z: np.ndarray,
        check_validity: bool = True,
        abs_tolerance: float = 1.0e-9,
    ) -> np.ndarray:
        """
        Returns a matrix representing the X,Y,Z transform of the reference frame.
        This reference frame is defined with Z being NADIR alignment, X facing
        the direction of the velocity, and Y as the cross product between Z and X
        complete the right handed system.

        The final matrix looks like such:
        [X1,Y1,Z1]
        [X2,Y2,Z2]
        [X3,Y3,Z3]

        Returns an array of matrices representing the reference frames for the given
        positions and velocities.

        Args:
            x (np.ndarray): position X (in any coordinate system)
            y (np.ndarray): position Y (in any coordinate system)
            z (np.ndarray): position Z (in any coordinate system)
            v_x (np.ndarray): velocity X (in any coordinate system)
            v_y (np.ndarray): velocity Y (in any coordinate system)
            v_z (np.ndarray): velocity Z (in any coordinate system)
            check_validity (bool, optional): assert that the dot product of the resulting rotation matrices are identities. Defaults to True.
            abs_tolerance (float, optional): tolerance value for the identity matrix validity check. Defaults to 1.e-2.

        Returns:
            np.ndarray: rotation matrix defining the reference frame.
        """

        # stack the vectors into a uniform format
        pos = np.column_stack([x, y, z])
        vel = np.column_stack([v_x, v_y, v_z])
        assert pos.ndim == vel.ndim

        r_norms = np.linalg.norm(pos, axis=1)
        v_norms = np.linalg.norm(vel, axis=1)
        z_axis = -pos / r_norms[:, np.newaxis]
        x_axis = vel / v_norms[:, np.newaxis]

        # Subtract the projection from the x-axis to make it orthogonal to the Z direction
        x_axis -= np.sum(x_axis * z_axis, axis=1)[:, np.newaxis] * z_axis
        x_axis /= np.linalg.norm(x_axis, axis=1)[:, np.newaxis]

        # define the y axis
        y_axis = np.cross(z_axis, x_axis, axis=1)

        # transpose the rotation matrix to make it 3x3xN
        rot_matrix = np.transpose(np.array([x_axis, y_axis, z_axis]), axes=(1, 2, 0))

        if check_validity:
            # Verify that each axis is orthogonal
            assert np.allclose(np.cross(x_axis, y_axis), z_axis, atol=abs_tolerance)
            assert np.allclose(np.cross(z_axis, x_axis), y_axis, atol=abs_tolerance)
            assert np.allclose(np.cross(y_axis, z_axis), x_axis, atol=abs_tolerance)

        return rot_matrix



from typing import List

import astropy.units as u
import numpy as np
import portion as P
from astropy.time import Time

from skywatch.time.interval import Interval

from .base_constraint import BaseAccessConstraint


class Access:
    def __init__(self, *constraints) -> None:
        """
        Access is used to calculate times when all provided constraints
        are satisfied.

        This class follows the builer pattern.

        Raises:
            TypeError: constraints must inherit from BaseAccessConstraint class.

        Args:
            consraints: BaseAccessConstraint objects representing the constraints
            that must be satisfied for access to return a non-empty Interval.
        """
        self.constraints = list()
        for constraint in constraints:
            if not isinstance(constraint, BaseAccessConstraint):
                raise TypeError(
                    "All constraints must inherit from BaseAccessConstraint class"
                )
            self.constraints.append(constraint)

        self._precision = 0.001 * u.s
        self._precise_endpoints = False
        self._min_duration = None
        self._max_duration = None
        self._only_check_failed_constraints = False

    def use_precise_endpoints(self, value: bool) -> "Access":
        """
        Toggle using high precision endpoints to get accurate
        start and stop times for access intervals.

        Args:
            value (bool): True or false

        Returns:
            Access: self
        """
        self._precise_endpoints = value
        return self

    def set_precision(self, precision: u.s = 0.001 * u.s) -> "Access":
        """
        Represents the delta time of the high precision constraint
        checks when *use_precise_endpoints* is True.

        Args:
            precision (u.s): delta time value in seconds.
            Default: 0.001 seconds.

        Returns:
            Access: self
        """
        self._precision = precision
        return self

    def set_min_duration(self, duration: u.s) -> "Access":
        """
        Sets the minimum duration of the resulting Interval
        when running access against all constraints. Successful
        access intervals must be greater than this duration to be
        included in the Interval output.

        Args:
            duration (u.s): minimum duration in seconds.

        Returns:
            Access: self
        """
        self._min_duration = duration
        return self

    def set_max_duration(self, duration: u.s) -> "Access":
        """
        Sets the maximum duration of the resulting Interval
        when running access against all constraints. Successful
        access intervals must be less than this duration to be
        included in the Interval output.

        Args:
            duration (u.s): maximum duration in seconds.

        Returns:
            Access: self
        """
        self._max_duration = duration
        return self

    def add_constraint(self, constraint: BaseAccessConstraint) -> "Access":
        """
        Add a constraint to this Access instance.

        Args:
            constraint (BaseAccessConstraint): Constraint that must be satisfied.

        Raises:
            TypeError: Must pass a subclass of BaseAccessConstraint.

        Returns:
            self: follows the builder pattern.
        """
        if not isinstance(constraint, BaseAccessConstraint):
            raise TypeError("Constraint must subclass BaseAccessConstraint.")

        self.constraints.append(constraint)
        return self

    def only_check_failed_constraints(self, value: bool) -> "Access":
        """
        When using precise endpoints, setting this to True will use an optimization
        to only re-check the constraints that have failed before the start and stop
        time of the access interval(s). If a constraint did not fail at the time period
        directly before or after the access interval, then it will not be recalculated.

        NOTE: If you are using a sufficiently large enough gap in your time intervals
        to check access during, then this has the potential to jump over a precise time
        that would have failed for a given constraint. However, if you know that your
        access constraints do not have a precision of less than the delta time between
        your time values, then this will not be an issue.

        NOTE: The defalt is set to False.

        Args:
            value (bool): True or False for whether or not to recheck only the failed
            constraints.

        Returns:
            Access: self
        """
        self._only_check_failed_constraints = bool(value)
        return self

    def calculate_at(self, time: Time, *args, **kwargs) -> Interval:
        """
        Calculates access using all access constraints in this Access instance
        at the specified time(s).

        Args:
            time (Time): Time(s) to check for calculate access intervals.

        Returns:
            Interval: Resulting TimeIntervals that represent time(s) when all the
            access constraints are successful.
        """
        return self.__call__(time, *args, **kwargs)

    def __call__(self, time: Time, *args, **kwargs) -> Interval:
        if not isinstance(time, Time):
            time = Time(time)

        # first pass check of access
        original_constrained_times = [np.array([True] * time.size)]
        for constraint in self.constraints:
            original_constrained_times.append(constraint(time, *args, **kwargs))

        # get all the windows from the first pass
        valid_time = np.all(original_constrained_times, axis=0)
        valid_ranges, access_times = Access._access_times(valid_time, time)

        # if no access or if not using precise timing, set the first pass as the final access
        if not self._precise_endpoints or len(self.constraints) == 0:
            final_access_times = access_times
            return self._compute_final_access_interval(final_access_times)

        # scale the precision to reflect the it in terms of seconds
        precision = (1 * u.s) / self._precision

        # calculate access at the precise time scale around the start and stop times of each window
        final_access_times = []
        for window_range, access_time in zip(valid_ranges, access_times):
            start_index = window_range[0]
            before_start_index = max(window_range[0] - 1, 0)

            if start_index == before_start_index:
                exact_start_time = time[
                    start_index
                ]  # nothing before the start time to interpolate to
            else:
                exact_start_time = self.__get_exact_times(
                    time,
                    start_index,
                    before_start_index,
                    precision,
                    original_constrained_times,
                    *args,
                    **kwargs
                )[0]

            # calculate the exact end time
            end_index = window_range[1] - 1
            after_end_index = min(window_range[1], time.size - 1)

            if end_index == after_end_index:
                exact_end_time = time[
                    end_index
                ]  # nothing after the end time to interpolate to
            else:
                exact_end_time = self.__get_exact_times(
                    time,
                    end_index,
                    after_end_index,
                    precision,
                    original_constrained_times,
                    *args,
                    **kwargs
                )[-1]

            final_access_times.append((exact_start_time, exact_end_time))

        return self._compute_final_access_interval(final_access_times)

    def __get_exact_times(
        self,
        time: Time,
        index_0: int,
        index_1: int,
        precision: u.Quantity,
        original_constrained_times: list,
        *args,
        **kwargs
    ) -> Time:
        """
        Helper to calculate the exact times between one index and another
        from the original times.

        Args:
            time (Time): original time values.
            index_0 (int): index of value 0
            index_1 (int): index of value 1
            precision (u.Quantity): precision in seconds
            original_constrained_times (list): boolean array of original contraint pass/fails

        Returns:
            Time: array of Time values.
        """
        # calculate number of steps between times to get desired precision
        t0 = time[index_0]
        t1 = time[index_1]
        num_steps = max(
            int(((t1 - t0).datetime.total_seconds()) * precision.value),
            2,
        )
        new_times = np.linspace(t0, t1, num_steps)

        # find the exact end time for this window
        constrained_times = [np.array([True] * len(new_times))]
        for index, constraint in enumerate(self.constraints):
            # check if the constraint needs to be computed again like above
            if (
                self._only_check_failed_constraints
                and original_constrained_times[index + 1][index_1]
            ):
                continue

            # need to get the precise time when this constraint turned False
            constrained_times.append(constraint(new_times, *args, **kwargs))

        return new_times[np.all(constrained_times, axis=0)]

    def _compute_final_access_interval(self, final_access_times: list):
        # compute the access windows using portion's intervals
        access = Interval()
        for access_time in final_access_times:
            t_start, t_end = access_time[0], access_time[-1]
            if self._min_duration is not None:
                if (
                    t_end - t_start
                ).datetime.total_seconds() * u.s < self._min_duration:
                    continue
            if self._max_duration is not None:
                if (
                    t_end - t_start
                ).datetime.total_seconds() * u.s > self._max_duration:
                    continue
            access = access.union(P.closed(t_start, t_end))
        return access

    @staticmethod
    def _get_true_ranges(bool_array: np.ndarray) -> List[tuple]:
        # Find the indices where the array changes from False to True or True to False
        change_indices = np.where(np.diff(bool_array))[0] + 1

        # If the first value is True, we need to add a 0 at the beginning
        if bool_array[0]:
            change_indices = np.insert(change_indices, 0, 0)

        # If the last value is True, we need to add the length of the array at the end
        if bool_array[-1]:
            change_indices = np.append(change_indices, len(bool_array))

        # Reshape the array into a 2D array where each row is a range of True values
        ranges = change_indices.reshape(-1, 2)

        # Convert the ranges to a list of tuples
        ranges = [tuple(r) for r in ranges]

        return ranges

    @staticmethod
    def _access_times(constrained_times: np.ndarray, time: Time) -> tuple:
        if not np.any(constrained_times):
            return [], []  # no access

        window_ranges = Access._get_true_ranges(constrained_times)
        valid_ranges = []
        access_times = []
        for range in window_ranges:
            new_time = time[range[0] : range[1]]
            if len(new_time) == 1:
                new_time = np.array([time[range[0]]] * 2)
            valid_ranges.append(range)
            access_times.append(new_time)

        return valid_ranges, access_times



import astropy.units as u
import pymap3d
from astropy.coordinates import ITRS, AltAz
from astropy.time import Time

from skywatch.attitude import BaseAttitudeStrategy
from skywatch.look_angles.aert import AzElRangeTime
from skywatch.look_angles.base_look_angle import BaseLookAngleStrategy
from skywatch.skypath.skypath import SkyPath


class LocalTangentENU(BaseLookAngleStrategy):
    def __init__(
        self,
        observer: SkyPath,
        attitude_strategy: BaseAttitudeStrategy = None,
        use_astropy: bool = False,
    ) -> None:
        """
        Using East, North, Up in reference to Earth. Up is zenith from the plane tangent to the earths
        surface at the observers location.

        Args:
            use_astropy (bool, optional): Do calculations using astropy's AltAz frame. NOTE: This is slower, but potentially
            more accurate. Defaults to False.
        """
        super().__init__()
        self.observer = observer
        self.use_astropy = use_astropy
        self.attitude_strategy = attitude_strategy

    def get_look_angles(
        self,
        target: SkyPath,
        time: Time,
    ) -> AzElRangeTime:
        if self.use_astropy:
            target_state = target.state_at(time, "itrs")
            alt_az = ITRS(
                target_state.cartesian.without_differentials(), obstime=time
            ).transform_to(
                AltAz(
                    location=self.observer.state_at(time, "itrs").earth_location,
                    obstime=time,
                )
            )
            az = alt_az.az
            el = alt_az.alt
            rng = alt_az.distance

        else:
            target_pos = target.state_at(time, "itrs").cartesian.xyz.to(u.m).value
            itrs_state = self.observer.state_at(time, "itrs")
            lats = itrs_state.earth_location.lat.value
            lons = itrs_state.earth_location.lon.value
            heights = itrs_state.earth_location.height.to(u.m).value

            enu = pymap3d.ecef2enu(*target_pos, lats, lons, heights)
            if self.attitude_strategy is not None:
                attitude = self.attitude_strategy.at(time)
                enu = attitude.inv().apply(enu)

            az, el, rng = pymap3d.enu2aer(*enu)
            az = az * u.deg
            el = el * u.deg
            rng = rng * u.m

        return AzElRangeTime(az, el, rng, time)



from setuptools import find_packages, setup

setup(
    name="skywatch",
    version="0.1.6",
    url="https://github.com/nsspencer/SkyWatch",
    author="Nathan Spencer",
    description="Aerospace/astrodynamics analysis library providing high level interfaces for coordinate, attitude, access, and look angle calculations.",
    packages=find_packages(),
    install_requires=["astropy", "numpy", "portion", "scipy", "pymap3d", "tqdm", "matplotlib", "pyvista", "basemap"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import astropy.units as u
import numpy as np
from astropy.time import Time
from scipy.spatial.transform import Rotation

from skywatch.attitude.base_attitude import BaseAttitudeStrategy
from skywatch.look_angles.aert import AzElRangeTime
from skywatch.look_angles.base_look_angle import BaseLookAngleStrategy
from skywatch.skypath.skypath import SkyPath


class DefaultLookStrategy(BaseLookAngleStrategy):
    def __init__(
        self,
        observer: SkyPath,
        attitude_strategy: BaseAttitudeStrategy,
        frame: str = "gcrs",
    ) -> None:
        """
        Represents a coordinate frame where X is facing the

        Args:
            observer (SkyPath): _description_
            attitude_strategy (BaseAttitudeStrategy): _description_
            frame (str, optional): _description_. Defaults to "gcrs".
        """
        super().__init__()
        self.observer = observer
        self.attitude_strategy = attitude_strategy
        self.frame = frame

    def get_look_angles(
        self,
        target: SkyPath,
        time: Time,
    ) -> AzElRangeTime:
        observer_position = (
            self.observer.state_at(time, self.frame).cartesian.xyz.to(u.m).value
        )
        target_position = target.state_at(time, self.frame).cartesian.xyz.to(u.m).value
        attitude = self.attitude_strategy.at(time)
        az, el, rng = DefaultLookStrategy.get_look_angles_to(
            observer_position.T, target_position.T, attitude
        )

        return AzElRangeTime(az * u.deg, el * u.deg, rng * u.m, time)

    @staticmethod
    def get_look_angles_to(
        observer_pos: np.ndarray, target_pos: np.ndarray, observer_frame: Rotation
    ) -> tuple:
        """
        Calculates the look angles from an observer with a given rotation to a target.

        Azimuth is defined as degrees clockwise.
            0 degrees faces the direction of velocity.
            90 degrees faces right of the direction of velocity.
            180 degrees faces opposite direction of velocity.
            270 degrees faces left of the direction of velocity.

        Elevation is defined as degrees of rise from the center of the observers frame.
            0 degrees faces the center of the observers frame (directly below the observer)
            90 degrees is in the direction of the velocity
            180 degrees faces the opposite direction of the center of the frame (directly above the observer)

        Range is defined as straight line (euclidean) distance from the observer to the target.

        Args:
            observer_pos (np.ndarray): 3 dimensional (X,Y,Z) array representing observers position.
            target_pos (np.ndarray): 3 dimensional (X,Y,Z) array representing targets position.
            observer_frame (Rotation): Attitude of the observer.

        Returns:
            tuple: azimuth, elevation, range
        """
        assert (
            observer_pos.ndim == target_pos.ndim
        ), "Dimensions of observer and target positions must be equivalent."
        if observer_pos.ndim == 1:
            observer_pos = np.reshape(observer_pos, (-1, 3))
            target_pos = np.reshape(target_pos, (-1, 3))

        # get the axis we need from the local frame
        _obs_frame_matrix = observer_frame.as_matrix()
        x_axis = _obs_frame_matrix[:, :, 0]
        z_axis = _obs_frame_matrix[:, :, 2]

        # Transform the target position to the local reference frame
        observer_to_target = target_pos - observer_pos

        # Project the observer_to_target vector onto the plane perpendicular to the z-axis
        component_along_z = (
            np.sum(observer_to_target * z_axis, axis=1)[:, np.newaxis] * z_axis
        )
        observer_to_target_projected = observer_to_target - component_along_z

        # calculate the azimuth from the x axis
        azimuth = DefaultLookStrategy._counterclockwise_angle_between(
            x_axis, observer_to_target_projected[:, :2]
        )

        # calculate the elevation angle
        elevation = DefaultLookStrategy._elevation_between(
            observer_to_target, component_along_z
        )

        # calculate the range
        rng = np.linalg.norm(observer_to_target, axis=1)

        return azimuth, elevation, rng

    @staticmethod
    def _elevation_between(v1: np.ndarray, v2: np.ndarray) -> np.ndarray:
        """Gets the elevation angle (in degrees) between the two vectors.

        Args:
            v1 (np.ndarray): vector 1
            v2 (np.ndarray): vector 2

        Returns:
            np.ndarray: elevation angle in degrees
        """
        # Normalize the vectors to unit vectors
        unit_vector1 = v1 / np.linalg.norm(v1, axis=1)[:, np.newaxis]
        unit_vector2 = v2 / np.linalg.norm(v2, axis=1)[:, np.newaxis]

        # Compute the dot product
        dot_product = np.sum((unit_vector1 * unit_vector2), axis=1)

        # Calculate the angle (in radians) & convert to degrees
        elevation = np.degrees(np.arccos(dot_product))
        return elevation

    @staticmethod
    def _counterclockwise_angle_between(v1: np.ndarray, v2: np.ndarray) -> np.ndarray:
        """Calculates the unambiguous counterclockwise angle between two vectors.

        Args:
            v1 (np.ndarray): vectors 1
            v2 (np.ndarray): vectors 2

        Returns:
            np.ndarray: angles in degrees (between 0 and 360)
        """
        angle_1 = np.arctan2(v1[:, 1], v1[:, 0]) * 180 / np.pi
        angle_2 = np.arctan2(v2[:, 1], v2[:, 0]) * 180 / np.pi
        angle = angle_1 - angle_2  # Swapped the order here to make counterclockwise
        angle = np.where(angle < 0, angle + 360, angle)
        return angle

    @staticmethod
    def _clockwise_angle_between(v1: np.ndarray, v2: np.ndarray) -> np.ndarray:
        """Calculates the unambiguous clockwise angle between two vectors.

        Args:
            v1 (np.ndarray): vectors 1
            v2 (np.ndarray): vectors 2

        Returns:
            np.ndarray: angles in degrees (between 0 and 360)
        """
        angle_1 = np.arctan2(v1[:, 1], v1[:, 0]) * 180 / np.pi
        angle_2 = np.arctan2(v2[:, 1], v2[:, 0]) * 180 / np.pi
        angle = angle_2 - angle_1
        angle = np.where(angle < 0, angle + 360, angle)
        return angle



import unittest

import astropy.units as u
import numpy as np
from astropy.time import Time
from scipy.spatial.transform import Rotation

from skywatch.attitude import LVLH, Fixed
from skywatch.look_angles import DefaultLookStrategy, LocalTangentENU
from skywatch.skypath import SkyPath

from .utils import get_ephem_as_skypath


def angle_differences(angles1, angles2):
    """
    Calculate the differences between corresponding angles in two arrays.
    """
    # Normalize angles to be between 0 and 360 degrees
    angles1 = angles1 % 360
    angles2 = angles2 % 360

    # Calculate the absolute difference between the angles
    diffs = np.abs(angles1 - angles2)

    return diffs


class Test1(unittest.TestCase):
    def test_ground_station_to_ground_station(self):
        gs1 = SkyPath.from_geodetic(Time("J2000"), 10 * u.deg, 10 * u.deg, 0 * u.m)
        gs2 = SkyPath.from_geodetic(Time("J2000"), 19 * u.deg, 1 * u.deg, 155000 * u.m)

        gs1_look_angles = LocalTangentENU(gs1)
        gs2_look_angles = LocalTangentENU(gs2)
        gs1_look_angles_offset = LocalTangentENU(
            gs1, Fixed(Rotation.from_euler("z", 10, degrees=True))
        )
        gs1_to_gs2 = gs1_look_angles.get_look_angles(gs2, Time("J2000"))
        gs2_to_gs1 = gs2_look_angles.get_look_angles(gs1, Time("J2000"))
        gs1_to_gs2_offset = gs1_look_angles_offset.get_look_angles(gs2, Time("J2000"))

        self.assertTrue(
            np.isclose(
                (gs1_to_gs2_offset.azimuth - gs1_to_gs2.azimuth).value, 10, atol=1e-7
            )
        )
        self.assertTrue(
            np.isclose(
                (gs1_to_gs2_offset.elevation - gs1_to_gs2.elevation).value, 0, atol=1e-7
            )
        )
        self.assertTrue(
            np.isclose((gs1_to_gs2_offset.range - gs1_to_gs2.range).value, 0, atol=1e-7)
        )

    def test_ground_station_to_satellite(self):
        sat = get_ephem_as_skypath()
        gs1 = SkyPath.from_geodetic(Time("J2000"), 10 * u.deg, 10 * u.deg, 0 * u.m)
        gs2 = SkyPath.from_geodetic(sat.obstime[0], 10 * u.deg, 10 * u.deg, 0 * u.m)

        gs_look_angles = LocalTangentENU(gs1)
        gs2sat = gs_look_angles.get_look_angles(sat, sat.obstime)

        gs2_look_angles = LocalTangentENU(gs2)
        gs2sat2 = gs2_look_angles.get_look_angles(sat, sat.obstime)

        self.assertTrue(np.allclose(gs2sat.azimuth, gs2sat2.azimuth))
        self.assertTrue(np.allclose(gs2sat.elevation, gs2sat2.elevation))
        self.assertTrue(np.allclose(gs2sat.range, gs2sat2.range))

    def test_satellite_to_ground_station(self):
        sat = get_ephem_as_skypath()
        ground_station = SkyPath.from_geodetic(
            sat.obstime, 34.5 * u.deg, -77.0 * u.deg, 0 * u.m
        )

        sat_look_angles_fn = DefaultLookStrategy(sat, LVLH(sat, "itrs"), "itrs")
        sat_look_angles = sat_look_angles_fn.get_look_angles(
            ground_station, sat.obstime
        )

        gs_look_angles_fn_gcrs = DefaultLookStrategy(sat, LVLH(sat, "gcrs"), "gcrs")
        sat_look_angles_gcrs = gs_look_angles_fn_gcrs.get_look_angles(
            ground_station, sat.obstime
        )

        self.assertTrue(
            np.min(
                angle_differences(
                    sat_look_angles.azimuth.value, sat_look_angles_gcrs.azimuth.value
                )
            )
            < 2.0
        )

        self.assertTrue(
            np.allclose(sat_look_angles.elevation, sat_look_angles_gcrs.elevation)
        )
        self.assertTrue(np.allclose(sat_look_angles.range, sat_look_angles_gcrs.range))

        sat_look_angles_offset_fn = DefaultLookStrategy(
            sat, LVLH(sat, "itrs", Rotation.from_euler("z", -45, degrees=True))
        )
        sat_2_gs_offset = sat_look_angles_offset_fn.get_look_angles(
            ground_station, sat.obstime
        )
        pass

    def test_look_angles(self):
        t_start = Time("2024-02-01T22:10:00")
        t_end = Time("2024-02-01T22:17:00")
        times = np.linspace(t_start, t_end, 8640)

        sat_position = get_ephem_as_skypath()
        ground_station_pos = SkyPath.from_geodetic(
            times, 34.5 * u.deg, -77.0 * u.deg, 0 * u.m
        )

        # gs_look_angles_fn = LocalTangentENU(ground_station_pos, False)
        # gs_look_angles_fn_astropy = LocalTangentENU(ground_station_pos, True)
        # gs_to_sat_look_angles = gs_look_angles_fn.get_look_angles(sat_position, times)
        # gs_to_sat_look_angles_astropy = gs_look_angles_fn_astropy.get_look_angles(
        #     sat_position, times
        # )

        sat_look_angles_fn = DefaultLookStrategy(sat_position, LVLH(sat_position))
        sat_look_angles_fn_offset = DefaultLookStrategy(
            sat_position,
            LVLH(
                sat_position,
                offset=Rotation.from_euler("z", -90, degrees=True),
            ),
        )
        sat_to_ground_station_look_angles = sat_look_angles_fn.get_look_angles(
            ground_station_pos, times
        )
        sat_to_ground_station_look_angles_offset = (
            sat_look_angles_fn_offset.get_look_angles(ground_station_pos, times)
        )

        extra_text = [
            f"norm: {az1}\noffset: {az2}"
            for az1, az2 in zip(
                sat_to_ground_station_look_angles.azimuth,
                sat_to_ground_station_look_angles_offset.azimuth,
            )
        ]

        plot_ground_track_and_points(
            times,
            sat_position,
            [
                ground_station_pos.itrs.earth_location.lat.value[0],
                ground_station_pos.itrs.earth_location.lon.value[0],
            ],
            extra_text=extra_text,
        )

        plot_look_angles(
            sat_to_ground_station_look_angles_offset.azimuth,
            sat_to_ground_station_look_angles_offset.elevation,
            sat_to_ground_station_look_angles_offset.range,
            sat_to_ground_station_look_angles_offset.time,
            "Offset plot",
        )

        plot_look_angles(
            sat_to_ground_station_look_angles.azimuth,
            sat_to_ground_station_look_angles.elevation,
            sat_to_ground_station_look_angles.range,
            sat_to_ground_station_look_angles.time,
            "LVLH plot",
        )


def plot_look_angles(
    azimuth: np.ndarray,
    elevation: np.ndarray,
    range: np.ndarray,
    times: np.ndarray,
    title: str = "Look Angles Plot",
) -> None:
    """Simple plotter to show the azimuth, elevation, and range look angles over time.

    Args:
        azimuth (np.ndarray): azimuth values
        elevation (np.ndarray): elevation values
        range (np.ndarray): range values
        times (np.ndarray): associated time values
    """
    import plotly.graph_objects as go

    times = [t.to_datetime() for t in times]
    # Create a trace for each line
    trace1 = go.Scatter(
        x=times, y=azimuth, mode="lines", name="Azimuth", line=dict(color="green")
    )
    trace2 = go.Scatter(
        x=times, y=elevation, mode="lines", name="Elevation", line=dict(color="blue")
    )
    trace3 = go.Scatter(
        x=times, y=range, mode="lines", name="Range", yaxis="y2", line=dict(color="red")
    )

    data = [trace1, trace2, trace3]

    layout = go.Layout(
        title=title,
        yaxis=dict(title="Angle", side="left"),
        yaxis2=dict(title="Distance", overlaying="y", side="right"),
        xaxis=dict(
            title="Date",
            showgrid=True,
            zeroline=True,
            showline=True,
            linecolor="rgb(204, 204, 204)",
            linewidth=2,
            showticklabels=True,
            tickformat="%H:%M",
        ),
    )

    fig = go.Figure(data=data, layout=layout)
    fig.show()


def plot_ground_track_and_points(
    time: Time,
    satellite: SkyPath,
    observation_llas: list = None,
    extra_text: list = None,
):
    import plotly.graph_objects as go

    # Create the line trace
    sat_earth_pos = satellite.state_at(time, "itrs").earth_location
    lats = sat_earth_pos.lat.value
    lons = sat_earth_pos.lon.value
    if extra_text is not None:
        text_values = [f"{str(t)}\n{str(extra)}" for t, extra in zip(time, extra_text)]
    else:
        text_values = [str(t) for t in time]

    ground_track = go.Scattergeo(
        lat=lats,
        lon=lons,
        mode="lines",
        line=dict(width=2, color="blue"),
        name="Line",
        text=text_values,
    )

    data = [ground_track]

    if observation_llas != None:
        if isinstance(observation_llas[0], list):
            observer_lats = [i[0] for i in observation_llas]
            observer_lons = [i[1] for i in observation_llas]
        else:
            observer_lats = [observation_llas[0]]
            observer_lons = [observation_llas[1]]
            print(
                f"plotting single point: [{observation_llas[0]}, {observation_llas[1]}]"
            )

        # Create the points trace
        observers = go.Scattergeo(
            lat=observer_lats,
            lon=observer_lons,
            mode="markers",
            marker=dict(size=20, color="red"),
            name="Points",
        )

        data.append(observers)

    # Create the plot
    fig = go.Figure(data=data)

    # Update the layout
    fig.update_layout(
        title_text="Line and Points on the Earth",
        showlegend=True,
        geo=dict(
            resolution=50,
            showland=True,
            showlakes=True,
            landcolor="rgb(204, 204, 204)",
            countrycolor="rgb(204, 204, 204)",
            lakecolor="rgb(255, 255, 255)",
            projection_type="equirectangular",
            coastlinewidth=2,
            lataxis=dict(range=[-90, 90], showgrid=True, dtick=10),
            lonaxis=dict(range=[-180, 180], showgrid=True, dtick=20),
        ),
    )

    fig.show()
